
## TODO: Remove once CSDownscale is on CRAN
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Interpolation.R')
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Intbc.R')
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Intlr.R')
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Analogs.R')
source('/esarchive/scratch/jramon/GitLab_jramon/downscaling/csdownscale/R/Utils.R')
source("https://earth.bsc.es/gitlab/external/cstools/-/raw/master/R/CST_BiasCorrection.R")


## Entry params data and recipe?
downscale_datasets <- function(data, recipe) {
  # Function that downscale the hindcast using the method stated in the
  # recipe. For the moment, forecast must be null.
  #
  # data: list of s2dv_cube objects containing the hcst, obs and fcst.
  # recipe: object obtained when passing the .yml recipe file to read_yaml()
  
  type <- tolower(recipe$Analysis$Workflow$Downscaling[["type"]])

  if (!is.null(data$fcst)) {
    warning("The downscaling will be only performed to the hindcast data")
    data$fcst <- NULL
  }

  if (type == "none") {
    
    hcst_calibrated <- data$hcst
    DOWNSCAL_MSG <- "##### NO DOWNSCALING PERFORMED #####"

  } else {
    # Downscaling function params
    int_method <- tolower(recipe$Analysis$Workflow$Downscaling[["int_method"]])
    bc_method <- tolower(recipe$Analysis$Workflow$Downscaling[["bc_method"]])
    lr_method <- tolower(recipe$Analysis$Workflow$Downscaling[["lr_method"]])
    target_grid <- tolower(recipe$Analysis$Workflow$Downscaling[["target_grid"]])
    nanalogs <- tolower(recipe$Analysis$Workflow$Downscaling[["nanalogs"]])

    if (is.null(recipe$Analysis$ncores)) {
      ncores <- 1
    } else {
      ncores <- recipe$Analysis$ncores
    }

    if (is.null(recipe$Analysis$loocv)) {
      loocv <- TRUE
    } else {
      loocv <- recipe$Analysis$loocv
    }

    if (is.null(recipe$Analysis$remove_NAs)) {
      na.rm = F
    } else {
      na.rm = recipe$Analysis$remove_NAs
    }
    
    CALIB_MSG <- "##### CALIBRATION COMPLETE #####"
    # Replicate observation array for the multi-model case
    if (mm) {
      obs.mm <- obs$data
      for(dat in 1:(dim(data$hcst$data)['dat'][[1]]-1)) {
        obs.mm <- abind(obs.mm, data$obs$data,
  		      along=which(names(dim(data$obs$data)) == 'dat'))
      }
      names(dim(obs.mm)) <- names(dim(obs$data))
      data$obs$data <- obs.mm
      remove(obs.mm)
    }
  
    if (recipe$Analysis$Variables$freq == "monthly_mean") {
    
      CST_CALIB_METHODS <- c("bias", "evmos", "mse_min", "crps_min", "rpc-based")
      ## TODO: implement other calibration methods
      ## TODO: Restructure the code?
      if (!(method %in% CST_CALIB_METHODS)) {
        stop("Calibration method in the recipe is not available for monthly",
  	   " data.")
      } else {
        ## Alba's version of CST_Calibration (pending merge) is being used
        # Calibrate the hindcast
        hcst_calibrated <- CST_Calibration(data$hcst, data$obs,
                                           cal.method = method,
                                           eval.method = "leave-one-out",
                                           multi.model = mm,
                                           na.fill = TRUE,
                                           na.rm = na.rm,
                                           apply_to = NULL,
                                           alpha = NULL,
                                           memb_dim = "ensemble",
                                           sdate_dim = "syear",
                                           ncores = ncores)
        if (!is.null(data$fcst)) {
          # Calibrate the forecast
          fcst_calibrated <- CST_Calibration(data$hcst, data$obs, data$fcst,
                                             cal.method = method,
                                             eval.method = "leave-one-out",
                                             multi.model = mm,
                                             na.fill = TRUE,
                                             na.rm = na.rm,
                                             apply_to = NULL,
                                             alpha = NULL,
                                             memb_dim = "ensemble",
                                             sdate_dim = "syear",
                                             ncores = ncores)
        } else {
          fcst_calibrated <- NULL
        }
      } 
    } else if (recipe$Analysis$Variables$freq == "daily_mean") {
      # Daily data calibration using Quantile Mapping
      if (!(method %in% c("qmap"))) {
        stop("Calibration method in the recipe is not available at daily ",
  	   "frequency. Only quantile mapping 'qmap' is implemented.")
      }
      # Calibrate the hindcast
      hcst_calibrated <- CST_QuantileMapping(data$hcst, data$obs,
        			             exp_cor = NULL,
        			             sample_dims = c("syear",
        						     "time",
        						     "ensemble"),
        				     sample_length = NULL,
        				     method = "QUANT",
					     wet.day = FALSE,
        				     ncores = ncores,
        				     na.rm = na.rm)
  
      if (!is.null(data$fcst)) {
        # Calibrate the forecast
        fcst_calibrated <- CST_QuantileMapping(data$hcst, data$obs,
        				       exp_cor = data$fcst,
        				       sample_dims = c("syear",
        						       "time",
        						       "ensemble"),
        				       sample_length = NULL,
        				       method = "QUANT",
					       wet.day = FALSE,
        				       ncores = ncores,
        				       na.rm = na.rm)
      } else {
        fcst_calibrated <- NULL
      }
    }
  }
print(CALIB_MSG)
  ## TODO: Return observations too?
  return(list(hcst = hcst_calibrated, fcst = fcst_calibrated))
}
